"""Used to send messages."""

import dataclasses
import json
import re
import typing
from urllib import error, parse, request

from tls.messaging.utils import notify_url as notify_url_util
from tls.messaging.utils import phone_number

from . import exceptions, oauth, types
from .utils import phone_number

_URL = "https://tapi.telstra.com/v2/messages/sms"


@dataclasses.dataclass
class TSms:
    """
    A sms.

    Attrs:
        to: The destination mobile number.
        delivery_status: Whether the delivery has been completed.
        message_id: Unique identifier for the message.
        message_status_url: URL to retrieve the current delivery status.

    """

    to: typing.Union[types.TTo, typing.List[types.TTo]]
    delivery_status: str
    message_id: types.TMessageId
    message_status_url: str


_VALID_FROM = re.compile(r"^[a-zA-Z0-9]+$")


def _send_validate_to(to: typing.Union[types.TTo, typing.List[types.TTo]]) -> None:
    """Validate the to parameter for send."""
    if not isinstance(to, str) and not isinstance(to, list):
        raise exceptions.SmsError(
            f'the value of "to" is not valid, expecting a string or a list of string, '
            f'received "{to}"'
        )
    if isinstance(to, str):
        result = phone_number.check(value=to)
        if not result.valid:
            raise exceptions.SmsError(
                f'the value of "to" is not valid, {result.reason}'
            )
    if isinstance(to, list):
        first_invalid_result = next(
            filter(lambda result: not result.valid, map(phone_number.check, to)),
            None,
        )
        if first_invalid_result is not None:
            raise exceptions.SmsError(
                f'the value of "to" is not valid, {first_invalid_result.reason}'
            )


def _send_validate_from(from_: typing.Optional[types.TFrom]) -> None:
    """Validate the from_ parameter for send."""
    if from_ is not None:
        if not isinstance(from_, str):
            raise exceptions.SmsError(
                'the value of "from_" is not valid, expected a string, received '
                f'"{from_}"'
            )
        if len(from_) > 11:
            raise exceptions.SmsError(
                'the value of "from_" has too many characters, expected at most 11 '
                f'characters, received "{from_}"'
            )
        if not _VALID_FROM.search(from_):
            raise exceptions.SmsError(
                'the value of "from_" contains invalid characters, expected alpha '
                f'numeric characters, received "{from_}"'
            )
        result = phone_number.check(value=from_)
        if result.valid:
            raise exceptions.SmsError(
                'the value of "from_" is a phone number, expected alpha '
                f'numeric characters that are not phone numbers, received "{from_}"'
            )


def _validate_send_args(  # pylint: disable=too-many-arguments
    to: typing.Union[types.TTo, typing.List[types.TTo]],
    body: types.TBody,
    from_: typing.Optional[types.TFrom],
    validity: typing.Optional[types.TValidity],
    scheduled_delivery: typing.Optional[types.TScheduledDelivery],
    notify_url: typing.Optional[types.TNotifyUrl],
    priority: typing.Optional[types.TPriority],
    reply_request: typing.Optional[types.TReplyRequest],
    receipt_off: typing.Optional[types.TReceiptOff],
    user_msg_ref: typing.Optional[types.TUsrMsgRef],
) -> None:
    """Validate the arguments for send."""
    # Validate to
    _send_validate_to(to)
    # Validate body
    if not isinstance(body, str):
        raise exceptions.SmsError(
            f'the value of "body" is not valid, expected a string, received "{body}"'
        )
    # Validate from_
    _send_validate_from(from_)
    # Validate validity
    if validity is not None and (
        not isinstance(validity, int) or isinstance(validity, bool)
    ):
        raise exceptions.SmsError(
            'the value of "validity" is not valid, expected an integer, received '
            f'"{validity}"'
        )
    # Validate scheduled_delivery
    if scheduled_delivery is not None and (
        not isinstance(scheduled_delivery, int) or isinstance(scheduled_delivery, bool)
    ):
        raise exceptions.SmsError(
            'the value of "scheduled_delivery" is not valid, expected an integer, '
            f'received "{scheduled_delivery}"'
        )
    # Validate notify_url
    notify_url_util.validate(value=notify_url, exception=exceptions.SmsError)
    # Validate priority
    if priority is not None and not isinstance(priority, bool):
        raise exceptions.SmsError(
            'the value of "priority" is not valid, expected an boolean, '
            f'received "{priority}"'
        )
    # Validate reply_request
    if reply_request is not None and not isinstance(reply_request, bool):
        raise exceptions.SmsError(
            'the value of "reply_request" is not valid, expected an boolean, '
            f'received "{reply_request}"'
        )
    # Validate receipt_off
    if receipt_off is not None and not isinstance(receipt_off, bool):
        raise exceptions.SmsError(
            'the value of "receipt_off" is not valid, expected an boolean, '
            f'received "{receipt_off}"'
        )
    # Validate user_msg_ref
    if user_msg_ref is not None and not isinstance(user_msg_ref, str):
        raise exceptions.SmsError(
            'the value of "user_msg_ref" is not valid, expected an string, '
            f'received "{user_msg_ref}"'
        )


def send(  # pylint: disable=too-many-arguments,too-many-locals
    to: typing.Union[types.TTo, typing.List[types.TTo]],
    body: types.TBody,
    from_: typing.Optional[types.TFrom] = None,
    validity: typing.Optional[types.TValidity] = None,
    scheduled_delivery: typing.Optional[types.TScheduledDelivery] = None,
    notify_url: typing.Optional[types.TNotifyUrl] = None,
    reply_request: typing.Optional[types.TReplyRequest] = None,
    priority: typing.Optional[types.TPriority] = None,
    receipt_off: typing.Optional[types.TReceiptOff] = None,
    user_msg_ref: typing.Optional[types.TUsrMsgRef] = None,
) -> TSms:
    """
    Send an SMS.

    Raises SmsError is anything goes wrong whilst sending a message.

    Args:
        to: The destination mobile number.
        body: The body of the message.
        from: Alpha numeric sender identity.
        validity: How long the platform should attempt to deliver the message for
            (in minutes).
        scheduled_delivery: How long the platform should wait before attempting to send
            the message (in minutes).
        notify_url: Contains a URL that will be called once your message has been
            processed.
        priority: Message will be placed ahead of all messages with a normal priority.
        reply_request: If set to true, the reply message functionality will be
            implemented.
        receipt_off: Whether Delivery Receipt will be sent back or not.
        user_msg_ref: Optional field used by some clients for custom reporting.

    Returns:
        The message that was sent.

    """
    _validate_send_args(
        to=to,
        body=body,
        from_=from_,
        validity=validity,
        scheduled_delivery=scheduled_delivery,
        notify_url=notify_url,
        priority=priority,
        reply_request=reply_request,
        receipt_off=receipt_off,
        user_msg_ref=user_msg_ref,
    )

    try:
        token = oauth.get_token()
    except exceptions.CredentialError as exc:
        raise exceptions.SmsError(f"Could not retrieve an OAuth token: {exc}") from exc

    data: typing.Dict[str, typing.Any] = {"to": to, "body": body}
    if from_ is not None:
        data["from"] = from_
    if validity is not None:
        data["validity"] = validity
    if scheduled_delivery is not None:
        data["scheduledDelivery"] = scheduled_delivery
    if notify_url is not None:
        data["notifyURL"] = notify_url
    if priority is not None:
        data["priority"] = priority
    if reply_request is not None:
        data["replyRequest"] = reply_request
    if receipt_off is not None:
        data["receiptOff"] = receipt_off
    if user_msg_ref is not None:
        data["userMsgRef"] = user_msg_ref
    data_str = json.dumps(data).encode()
    headers = {
        "Content-Type": "application/json",
        "Authorization": token.authorization,
    }
    sms_request = request.Request(_URL, data=data_str, headers=headers, method="POST")
    try:
        with request.urlopen(sms_request) as response:
            sms_dict = json.loads(response.read().decode())
            return TSms(
                to=sms_dict["messages"][0]["to"],
                delivery_status=sms_dict["messages"][0]["deliveryStatus"],
                message_id=sms_dict["messages"][0]["messageId"],
                message_status_url=sms_dict["messages"][0]["messageStatusURL"],
            )
    except error.HTTPError as exc:
        raise exceptions.SmsError(f"Could not send SMS: {exc}") from exc


@dataclasses.dataclass
class TReply:
    """
    A reply.

    Attrs:
        destination_address: Where the message is delivered to.
        sender_address: Who the message is from.
        status: Whether the delivery has been completed.
        message: The body of the message.
        message_id: Unique identifier for the message.
        sent_timestamp: When the message was sent.

    """

    destination_address: str
    sender_address: str
    status: str
    message: str
    message_id: types.TMessageId
    sent_timestamp: types.TSentTimestamp


def get_next_unread_reply() -> typing.Optional[TReply]:
    """
    Get the next unread reply that have been received.

    Raises SmsError is anything goes wrong whilst retrieving the reply.

    Returns:
        The next unread reply or None if there are no more replies.

    """
    try:
        token = oauth.get_token()
    except exceptions.CredentialError as exc:
        raise exceptions.SmsError(f"Could not retrieve an OAuth token: {exc}") from exc

    headers = {"Authorization": token.authorization}
    reply_request = request.Request(_URL, headers=headers, method="GET")
    try:
        with request.urlopen(reply_request) as response:
            reply_dict = json.loads(response.read().decode())
            try:
                return TReply(
                    status=reply_dict["status"],
                    destination_address=reply_dict["destinationAddress"],
                    sender_address=reply_dict["senderAddress"],
                    message_id=reply_dict["messageId"],
                    message=reply_dict["message"],
                    sent_timestamp=reply_dict["sentTimestamp"],
                )
            except KeyError:
                return None
    except error.HTTPError as exc:
        raise exceptions.SmsError(f"Could not send SMS: {exc}") from exc


@dataclasses.dataclass
class TStatus:
    """
    The status of a message.

    Attrs:
        to: Where the message is delivered to.
        delivery_status: Whether the delivery has been completed.
        received_timestamp: When the message was received.
        sent_timestamp: When the message was sent.

    """

    to: types.TTo
    sent_timestamp: types.TSentTimestamp
    received_timestamp: str
    delivery_status: str


def get_status(message_id: types.TMessageId) -> TStatus:
    """
    Retrieve the status of a message.

    Raises SmsError is anything goes wrong whilst retrieving the status.

    Args:
        message_id: Unique identifier for the message.

    Returns:
        The status of the message.

    """
    try:
        token = oauth.get_token()
    except exceptions.CredentialError as exc:
        raise exceptions.SmsError(f"Could not retrieve an OAuth token: {exc}") from exc

    headers = {"Authorization": token.authorization, "Content-Type": "application/json"}
    status_request = request.Request(
        f"{_URL}/{parse.quote(message_id)}/status", headers=headers, method="GET"
    )
    try:
        with request.urlopen(status_request) as response:
            status_dict = json.loads(response.read().decode())
            return TStatus(
                to=status_dict[0]["to"],
                sent_timestamp=status_dict[0]["sentTimestamp"],
                received_timestamp=status_dict[0]["receivedTimestamp"],
                delivery_status=status_dict[0]["deliveryStatus"],
            )
    except error.HTTPError as exc:
        raise exceptions.SmsError(f"Could not retrieve the status: {exc}") from exc



"""Shared types for messaging."""

TTo = str
TFrom = str
TNotifyUrl = str
TMessageId = str
TBody = str
TSentTimestamp = str
TValidity = int
TScheduledDelivery = int
TReplyRequest = bool
TPriority = bool
TReceiptOff = bool
TUsrMsgRef = str



"""Utilities for the notify URL."""

import typing

from .. import exceptions, types


def validate(
    *,
    value: typing.Optional[types.TNotifyUrl],
    exception: typing.Type[exceptions.SmsBaseException],
) -> None:
    """Validate the notify_url parameter for send."""
    if value is not None:
        if not isinstance(value, str):
            raise exception(
                'the value of "notify_url" is not valid, expected a string, received '
                f'"{value}"'
            )
        if not value.lower().startswith("https"):
            raise exception(
                'the value of "notify_url" is not valid, it must start with https, '
                f'received "{value}"'
            )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tls', 'tls.messaging', 'tls.messaging.utils']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'tls.messaging',
    'version': '0.6.1',
    'description': 'SDK for the Telstra Messaging API',
    'long_description': '# Telstra Messaging\n\nThe SDK for the Telstra messaging API.\n\n> :warning: **This SDK is experimental, everything is subject to change**\n\n## Installing\n\n```bash\npip install tls.messaging\n```\n\n## Getting Started\n\nSet the `TLS_CLIENT_KEY` and `TLS_CLIENT_SECRET` environment variables. These\nare the `Client key` and `Client secret` you can find here:\n<https://dev.telstra.com/user/me/apps>.\n\nTo send your first SMS:\n\n```python\nfrom tls.messaging import sms\n\nsms.send(to="+61412345678", body="Hi")\n```\n\nTo set the required environment variables if your application is in `app.py`:\n\n```bash\nTLS_CLIENT_KEY="<client key>" TLS_CLIENT_SECRET="<client secret>" python app.py\n```\n\n## Authentication\n\nOn top of the authentication through the `TLS_CLIENT_KEY` and\n`TLS_CLIENT_SECRET` environment variables, authentication through code is also\nsupported. For example:\n\n```python\nfrom tls.messaging.utils.config import CONFIG\n\nCONFIG.tls_client_key = \'<client key>\'\nCONFIG.tls_client_secret = \'<client secret>\'\n```\n\nThis should be done before any interactions requiring authentication, such as\nsending a SMS.\n\n## Free Trial\n\nTelstra offers a free trial for the messaging API to help you evaluate whether\nit meets your needs. There are some restrictions that apply compared to the\nfull API, including a maximum number of SMS that can be sent and requiring the\nregistration of a limited number of destinations before SMS can be sent to that\ndestination. For more information, please see here:\n<https://dev.telstra.com/content/messaging-api#tag/Free-Trial>.\n\n### Registering Destinations\n\n> :information_source: **Only required for the free trial**\n\nRegister destinations for the free trial. For more information, please see\nhere:\n<https://dev.telstra.com/content/messaging-api#operation/freeTrialBnumRegister>.\n\nThe function `tls.messaging.bnum.register` can be used to register\ndestinations. It takes the following arguments:\n\n- `phone_numbers`: A list of destinations, expected to be phone numbers of the\n  form `+614XXXXXXXX` or `04XXXXXXXX`.\n\nIt returns the list of phone numbers that have been registered.\n\nFor example:\n\n```python\nfrom tls.messaging import bnum\n\nphone_numbers = bnum.register(phone_numbers=["+61412345678"])\nprint(phone_numbers)\n```\n\n### Retrieve Destinations\n\n> :information_source: **Only required for the free trial**\n\nRetrieve destinations for the free trial. For more information, please see\nhere:\n<https://dev.telstra.com/content/messaging-api#operation/freeTrialBnumList>.\n\nThe function `tls.messaging.bnum.get` can be used to register\ndestinations. It takes no arguments. It returns the list of phone numbers that\nhave been registered.\n\nFor example:\n\n```python\nfrom tls.messaging import bnum\n\nphone_numbers = bnum.get()\nprint(phone_numbers)\n```\n\n## Subscription\n\nA subscription gives you a dedicated mobile number tied to an application. For\nmore information, please see here:\n<https://dev.telstra.com/content/messaging-api#tag/Provisioning>.\n\n### Create Subscription\n\nCreate a new subscription. For more information, please see here:\n<https://dev.telstra.com/content/messaging-api#operation/createSubscription>.\n\nThe function `tls.messaging.subscription.create` can be used to create a\nsubscription. It takes the following arguments:\n\n- `active_days` (optional): The number of days the subscription will be active,\n  defaults to 30.\n- `notify_url` (optional): A notification URL that will be POSTed to whenever a\n  new message (i.e. a reply to a message sent) arrives at this destination\n  address.\n\nIt returns an object with the following properties:\n\n- `destination_address`: The phone number that a message can be sent to.\n- `active_days`: The number of days left on the subscription.\n\nFor example:\n\n```python\nfrom tls.messaging import subscription\n\ncreated_subscription = subscription.create()\nprint(created_subscription)\n```\n\n### Get Subscription\n\nRetrieve the current subscription. For more information, please see here:\n<https://dev.telstra.com/content/messaging-api#operation/getSubscription>.\n\nThe function `tls.messaging.subscription.get` can be used to get the current\nsubscription. It takes no arguments. It returns an object with the following\nproperties:\n\n- `destination_address`: The phone number that a message can be sent to.\n- `active_days`: The number of days left on the subscription.\n\nFor example:\n\n```python\nfrom tls.messaging import subscription\n\nretrieved_subscription = subscription.get()\nprint(retrieved_subscription)\n```\n\n### Delete Subscription\n\nDelete the current subscription. For more information, please see here:\n<https://dev.telstra.com/content/messaging-api#operation/deleteSubscription>.\n\nThe function `tls.messaging.subscription.delete` can be used to delete the current\nsubscription. It takes no arguments.\n\n```python\nfrom tls.messaging import subscription\n\nsubscription.delete()\n```\n\n## SMS\n\nFor more information, please see here:\n<https://dev.telstra.com/content/messaging-api#tag/Messaging>.\n\n### Send SMS\n\nSend a SMS to a mobile number. For more information, please see here:\n<https://dev.telstra.com/content/messaging-api#operation/sendSms>.\n\nThe function `tls.messaging.sms.send` can be used to send SMS. It takes the\nfollowing arguments:\n\n- `to`: The destination address, expected to be a phone number of the form\n  `+614XXXXXXXX` or `04XXXXXXXX`.\n- `body`: The SMS to send.\n- `from_` (optional): An alphanumeric value which will appear as the sender.\n  Note that phone numbers are not supported amd the maximum length is 11\n  characters. Certain well know senders will be blocked.\n- `validity` (optional): How long the platform should attempt to deliver the\n  message for (in minutes).\n- `scheduled_delivery` (optional): How long the platform should wait before\n  attempting to send the message (in minutes).\n- `notify_url` (optional): Contains a URL that will be called once your message\n  has been processed.\n- `priority` (optional): Message will be placed ahead of all messages with a\n  normal priority.\n- `reply_request` (optional): If set to true, the reply message functionality\n  will be implemented.\n- `receipt_off` (optional): Whether Delivery Receipt will be sent back or not.\n- `user_msg_ref` (optional): Optional field used by some clients for custom\n  reporting.\n\nIt returns an object with the following properties:\n\n- `to`: The destination mobile number.\n- `delivery_status`: Whether the delivery has been completed.\n- `message_id`: Unique identifier for the message.\n- `message_status_url`: URL to retrieve the current delivery status.\n\nFor example:\n\n```python\nfrom tls.messaging import sms\n\nsms.send(to="+61412345678", body="Hi")\n```\n\n### Get SMS Status\n\nFind out whether a SMS has been sent. For more information, please see here:\n<https://dev.telstra.com/content/messaging-api#operation/getSmsStatus>.\n\nThe function `tls.messaging.sms.get_status` can be used to retrieve\nthe status of a SMS. It takes the following arguments:\n\n- `message_id`:Unique identifier for the message.\n\nIt returns an object with the following properties:\n\n- `to`: Where the message is delivered to.\n- `delivery_status`: Whether the delivery has been completed.\n- `received_timestamp`: When the message was received.\n- `sent_timestamp`: When the message was sent.\n\nFor example:\n\n```python\nfrom tls.messaging import sms\n\nsent_sms = sms.send(to="+61412345678", body="Hi")\nstatus = sms.get_status(sent_sms.message_id)\nprint(status)\n```\n\n### Retrieve Replies\n\nRetrieve SMS sent to the mobile number associated with the subscription. For\nmore information, please see here:\n<https://dev.telstra.com/content/messaging-api#operation/retrieveSmsReplies>.\n\nThe function `tls.messaging.sms.get_next_unread_reply` can be used to retrieve\nthe next unread reply for your phone number subscription. It takes no\narguments. It returns `None` if there are no more replies or an object with the\nfollowing properties:\n\n- `destination_address`: Where the message is delivered to.\n- `sender_address`: Who the message is from.\n- `status`: Whether the delivery has been completed.\n- `message`: The body of the message.\n- `message_id`: Unique identifier for the message.\n- `sent_timestamp`: When the message was sent.\n\nFor example:\n\n```python\nfrom tls.messaging import sms\n\nreply = sms.get_next_unread_reply()\nprint(reply)\n```\n',
    'author': 'David Andersson',
    'author_email': 'david-andersson@users.noreply.github.com ',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
